@@ -60,7 +60,7 @@ garbage_collect (qxl_screen_t *qxl)
 	     */
 #define POINTER_MASK ((1 << 2) - 1)
 	    
-	    union qxl_release_info *info = (void *)(id & ~POINTER_MASK);
+	    union qxl_release_info *info = u64_to_pointer (id & ~POINTER_MASK);
 	    struct qxl_cursor_cmd *cmd = (struct qxl_cursor_cmd *)info;
 	    struct qxl_drawable *drawable = (struct qxl_drawable *)info;
 	    int is_cursor = FALSE;
@@ -71,14 +71,14 @@ garbage_collect (qxl_screen_t *qxl)
 	    if (is_cursor && cmd->type == QXL_CURSOR_SET)
 	    {
 		struct qxl_cursor *cursor = (void *)virtual_address (
-		    qxl, (void *)cmd->u.set.shape);
+		    qxl, u64_to_pointer (cmd->u.set.shape));
 
 		qxl_free (qxl->mem, cursor);
 	    }
 	    else if (!is_cursor && drawable->type == QXL_DRAW_COPY)
 	    {
 		struct qxl_image *image = virtual_address (
-		    qxl, (void *)drawable->u.copy.src_bitmap);
+		    qxl, u64_to_pointer (drawable->u.copy.src_bitmap));
 
 		qxl_image_destroy (qxl, image);
 	    }
@@ -196,6 +196,9 @@ qxl_unmap_memory(qxl_screen_t *qxl, int scrnIndex)
 #endif
 
     qxl->ram = qxl->ram_physical = qxl->vram = qxl->rom = NULL;
+
+    qxl->num_modes = 0;
+    qxl->modes = NULL;
 }
 
 static Bool
@@ -206,7 +209,7 @@ qxl_map_memory(qxl_screen_t *qxl, int scrnIndex)
 			 qxl->pci->regions[0].size,
 			 PCI_DEV_MAP_FLAG_WRITABLE | PCI_DEV_MAP_FLAG_WRITE_COMBINE,
 			 &qxl->ram);
-    qxl->ram_physical = (void *)qxl->pci->regions[0].base_addr;
+    qxl->ram_physical = u64_to_pointer (qxl->pci->regions[0].base_addr);
 
     pci_device_map_range(qxl->pci, qxl->pci->regions[1].base_addr, 
 			 qxl->pci->regions[1].size,
@@ -215,7 +218,7 @@ qxl_map_memory(qxl_screen_t *qxl, int scrnIndex)
 
     pci_device_map_range(qxl->pci, qxl->pci->regions[2].base_addr, 
 			 qxl->pci->regions[2].size, 0,
-			 &qxl->rom);
+			 (void **)&qxl->rom);
 
     qxl->io_base = qxl->pci->regions[3].base_addr;
 #else
@@ -240,6 +243,9 @@ qxl_map_memory(qxl_screen_t *qxl, int scrnIndex)
     xf86DrvMsg(scrnIndex, X_INFO, "ram at %p; vram at %p; rom at %p\n",
 	       qxl->ram, qxl->vram, qxl->rom);
 
+    qxl->num_modes = *(uint32_t *)((uint8_t *)qxl->rom + qxl->rom->modes_offset);
+    qxl->modes = (struct qxl_mode *)(((uint8_t *)qxl->rom) + qxl->rom->modes_offset + 4);
+
     return TRUE;
 }
 
@@ -265,15 +271,16 @@ static Bool
 qxl_switch_mode(int scrnIndex, DisplayModePtr p, int flags)
 {
     qxl_screen_t *qxl = xf86Screens[scrnIndex]->driverPrivate;
-    struct qxl_mode *m = (void *)p->Private;
+    int mode_index = (int)(unsigned long)p->Private;
+    struct qxl_mode *m = qxl->modes + mode_index;
     ScreenPtr pScreen = qxl->pScrn->pScreen;
 
     if (!m)
 	return FALSE;
 
     /* if (debug) */
-    xf86DrvMsg(scrnIndex, X_INFO, "Setting mode %d (%d x %d) (%d x %d) %p\n",
-	       m->id, m->x_res, m->y_res, p->HDisplay, p->VDisplay, p);
+    xf86DrvMsg (scrnIndex, X_INFO, "Setting mode %d (%d x %d) (%d x %d) %p\n",
+		m->id, m->x_res, m->y_res, p->HDisplay, p->VDisplay, p);
 
     outb(qxl->io_base + QXL_IO_RESET, 0);
     
@@ -348,11 +355,11 @@ make_drawable (qxl_screen_t *qxl, uint8_t type,
 
     CHECK_POINT();
 
-    drawable->release_info.id = (uint64_t)drawable;
+    drawable->release_info.id = pointer_to_u64 (drawable);
 
     drawable->type = type;
 
-    drawable->effect = QXL_EFFECT_OPAQUE;
+    drawable->effect = QXL_EFFECT_BLEND;
     drawable->bitmap_offset = 0;
     drawable->bitmap_area.top = 0;
     drawable->bitmap_area.left = 0;
@@ -509,13 +516,13 @@ qxl_send_copies (qxl_screen_t *qxl)
     nbox = REGION_NUM_RECTS (&qxl->to_be_sent);
     pBox = REGION_RECTS (&qxl->to_be_sent);
 
-/*     print_region ("send bits", &qxl->to_be_sent); */
-/*     print_region ("unsent bits", &qxl->pending_copy); */
+/*      if (REGION_NUM_RECTS (&qxl->to_be_sent) > 0)  */
+/*        	print_region ("send bits", &qxl->to_be_sent); */
     
     while (nbox--)
     {
 	struct qxl_rect qrect;
-	
+
 	qrect.top = pBox->y1;
 	qrect.left = pBox->x1;
 	qrect.bottom = pBox->y2;
@@ -542,7 +549,8 @@ paint_shadow (qxl_screen_t *qxl)
     submit_copy (qxl, &qrect);
 }
 
-static void qxl_sanity_check(qxl_screen_t *qxl)
+static void
+qxl_sanity_check (qxl_screen_t *qxl)
 {
     /* read the mode back from the rom */
     if (!qxl->rom || !qxl->pScrn)
@@ -556,7 +564,7 @@ static void qxl_sanity_check(qxl_screen_t *qxl)
 }
 
 static void
-qxl_block_handler(pointer data, OSTimePtr pTimeout, pointer pRead)
+qxl_block_handler (pointer data, OSTimePtr pTimeout, pointer pRead)
 {
     qxl_screen_t *qxl = (qxl_screen_t *) data;
 
@@ -568,7 +576,7 @@ qxl_block_handler(pointer data, OSTimePtr pTimeout, pointer pRead)
 }
 
 static void
-qxl_wakeup_handler(pointer data, int i, pointer LastSelectMask)
+qxl_wakeup_handler (pointer data, int i, pointer LastSelectMask)
 {
 }
 
@@ -602,9 +610,7 @@ qxl_on_damage (DamagePtr pDamage, RegionPtr pRegion, pointer closure)
 
 /*     print_region ("damage", pRegion); */
     
-#if 0
-    ErrorF ("damage\n");
-#endif
+/*     print_region ("on_damage ", pRegion); */
 
     accept_damage (qxl);
 
@@ -641,7 +647,6 @@ qxl_create_screen_resources(ScreenPtr pScreen)
 
     REGION_INIT (pScreen, &(qxl->pending_copy), NullBox, 0);
 
-    ErrorF ("initialized\n");
     REGION_INIT (pScreen, &(qxl->to_be_sent), NullBox, 0);
  
     DamageRegister (&pPixmap->drawable, qxl->damage);
@@ -742,6 +747,33 @@ qxl_copy_n_to_n (DrawablePtr    pSrcDrawable,
 
 /* 	ErrorF ("Accelerated copy: %d boxes\n", n); */
 
+	/* At this point we know that any pending damage must
+	 * have been caused by whatever copy operation triggered us.
+	 * 
+	 * Therefore we can clear it.
+	 *
+	 * We couldn't clear it at the toplevel function because 
+	 * the copy might end up being empty, in which case no
+	 * damage would have been generated. Which means the
+	 * pending damage would have been caused by some
+	 * earlier operation.
+	 */
+	if (n)
+	{
+/* 	    ErrorF ("Clearing pending damage\n"); */
+	    clear_pending_damage (qxl);
+	    
+	    /* We have to do this because the copy will cause the damage
+	     * to be sent to move.
+	     * 
+	     * Instead of just sending the bits, we could also move
+	     * the existing damage around; however that's a bit more 
+	     * complex, and the performance win is unlikely to be
+	     * very big.
+	     */
+	    qxl_send_copies (qxl);
+	}
+    
 	while (n--)
 	{
 	    struct qxl_drawable *drawable;
@@ -784,22 +816,14 @@ static RegionPtr
 qxl_copy_area(DrawablePtr pSrcDrawable, DrawablePtr pDstDrawable, GCPtr pGC,
 	    int srcx, int srcy, int width, int height, int dstx, int dsty)
 {
-    ScreenPtr pScreen = pSrcDrawable->pScreen;
-    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
-    qxl_screen_t *qxl = pScrn->driverPrivate;
-
     if (pSrcDrawable->type == DRAWABLE_WINDOW &&
 	pDstDrawable->type == DRAWABLE_WINDOW)
     {
 	RegionPtr res;
 
-	/* We have to do this because the copy will cause the damage
-	 * to be sent to move.
-	 */
-	clear_pending_damage (qxl);
+/* 	ErrorF ("accelerated copy %d %d %d %d %d %d\n",  */
+/* 		srcx, srcy, width, height, dstx, dsty); */
 
-	qxl_send_copies (qxl);
-    
 	res = fbDoCopy (pSrcDrawable, pDstDrawable, pGC,
 			srcx, srcy, width, height, dstx, dsty,
 			qxl_copy_n_to_n, 0, NULL);
@@ -808,6 +832,9 @@ qxl_copy_area(DrawablePtr pSrcDrawable, DrawablePtr pDstDrawable, GCPtr pGC,
     }
     else
     {
+/* 	ErrorF ("Falling back %d %d %d %d %d %d\n",  */
+/* 		srcx, srcy, width, height, dstx, dsty); */
+
 	return fbCopyArea (pSrcDrawable, pDstDrawable, pGC,
 			   srcx, srcy, width, height, dstx, dsty);
     }
@@ -870,20 +897,9 @@ qxl_paint_window(WindowPtr pWin, RegionPtr pRegion, int what)
 static void
 qxl_copy_window (WindowPtr pWin, DDXPointRec ptOldOrg, RegionPtr prgnSrc)
 {
-    ScreenPtr pScreen = pWin->drawable.pScreen;
-    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
-    qxl_screen_t *qxl = pScrn->driverPrivate;
     RegionRec rgnDst;
     int dx, dy;
 
-    /* We have to do this because the copy will cause the damage
-     * to be sent to move, which means the "undamage" later will
-     * remove the wrong damage.
-     */
-    clear_pending_damage (qxl);
-
-    qxl_send_copies (qxl);
-
     dx = ptOldOrg.x - pWin->drawable.x;
     dy = ptOldOrg.y - pWin->drawable.y;
 
@@ -933,6 +949,7 @@ qxl_screen_init(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
     qxl_screen_t *qxl = pScrn->driverPrivate;
     struct qxl_rom *rom;
     struct qxl_ram_header *ram_header;
+    VisualPtr visual;
 
     CHECK_POINT();
 
@@ -966,18 +983,21 @@ qxl_screen_init(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 		      pScrn->currentMode->VDisplay,
 		      pScrn->xDpi, pScrn->yDpi, pScrn->displayWidth,
 		      pScrn->bitsPerPixel))
-	goto out;
     {
-	VisualPtr visual = pScreen->visuals + pScreen->numVisuals;
-	while (--visual >= pScreen->visuals) {
-	    if ((visual->class | DynamicClass) == DirectColor) {
-                visual->offsetRed = pScrn->offset.red;
-                visual->offsetGreen = pScrn->offset.green;
-                visual->offsetBlue = pScrn->offset.blue;
-                visual->redMask = pScrn->mask.red;
-                visual->greenMask = pScrn->mask.green;
-                visual->blueMask = pScrn->mask.blue;
-	    }
+	goto out;
+    }
+
+    visual = pScreen->visuals + pScreen->numVisuals;
+    while (--visual >= pScreen->visuals) 
+    {
+	if ((visual->class | DynamicClass) == DirectColor) 
+	{
+	    visual->offsetRed = pScrn->offset.red;
+	    visual->offsetGreen = pScrn->offset.green;
+	    visual->offsetBlue = pScrn->offset.blue;
+	    visual->redMask = pScrn->mask.red;
+	    visual->greenMask = pScrn->mask.green;
+	    visual->blueMask = pScrn->mask.blue;
 	}
     }
 
@@ -1071,7 +1091,9 @@ qxl_color_setup(ScrnInfoPtr pScrn)
 
     if (!xf86SetDepthBpp(pScrn, 0, 0, 0, Support32bppFb))
 	return FALSE;
-    if (pScrn->depth != 16 && pScrn->depth != 24) {
+
+    if (pScrn->depth != 15 && pScrn->depth != 24) 
+    {
 	xf86DrvMsg(scrnIndex, X_ERROR, "Depth %d is not supported\n",
 		   pScrn->depth);
 	return FALSE;
@@ -1080,8 +1102,10 @@ qxl_color_setup(ScrnInfoPtr pScrn)
 
     if (!xf86SetWeight(pScrn, rzeros, rzeros))
 	return FALSE;
+
     if (!xf86SetDefaultVisual(pScrn, -1))
 	return FALSE;
+
     if (!xf86SetGamma(pScrn, gzeros))
 	return FALSE;
 
@@ -1089,23 +1113,28 @@ qxl_color_setup(ScrnInfoPtr pScrn)
 }
 
 static void
-qxl_print_mode(int scrnIndex, void *p)
+print_modes (qxl_screen_t *qxl, int scrnIndex)
 {
-    struct qxl_mode *m = p;
-    xf86DrvMsg(scrnIndex, X_INFO,
-	       "%d: %dx%d, %d bits, stride %d, %dmm x %dmm, orientation %d\n",
-	       m->id, m->x_res, m->y_res, m->bits, m->stride, m->x_mili,
-	       m->y_mili, m->orientation);
+    int i;
+
+    for (i = 0; i < qxl->num_modes; ++i)
+    {
+	struct qxl_mode *m = qxl->modes + i;
+
+	xf86DrvMsg (scrnIndex, X_INFO,
+		    "%d: %dx%d, %d bits, stride %d, %dmm x %dmm, orientation %d\n",
+		    m->id, m->x_res, m->y_res, m->bits, m->stride, m->x_mili,
+		    m->y_mili, m->orientation);
+    }
 }
 
 static Bool
 qxl_check_device(ScrnInfoPtr pScrn, qxl_screen_t *qxl)
 {
     int scrnIndex = pScrn->scrnIndex;
-    int i, mode_offset;
     struct qxl_rom *rom = qxl->rom;
     struct qxl_ram_header *ram_header = (void *)((unsigned long)qxl->ram + rom->ram_header_offset);
-    
+
     CHECK_POINT();
     
     if (rom->magic != 0x4f525851) { /* "QXRO" little-endian */
@@ -1145,17 +1174,10 @@ qxl_check_device(ScrnInfoPtr pScrn, qxl_screen_t *qxl)
     qxl->draw_area_size = rom->draw_area_size;
     pScrn->videoRam = rom->draw_area_size / 1024;
     
-    mode_offset = rom->modes_offset / 4;
-    qxl->num_modes = ((uint32_t *)rom)[mode_offset];
-    xf86DrvMsg(scrnIndex, X_INFO, "%d available modes:\n", qxl->num_modes);
-    qxl->modes = (void *)((uint32_t *)rom + mode_offset + 1);
-    for (i = 0; i < qxl->num_modes; i++)
-	qxl_print_mode(scrnIndex, qxl->modes + i);
-
     return TRUE;
 }
 
-static struct qxl_mode *
+static int
 qxl_find_native_mode(ScrnInfoPtr pScrn, DisplayModePtr p)
 {
     int i;
@@ -1163,17 +1185,28 @@ qxl_find_native_mode(ScrnInfoPtr pScrn, DisplayModePtr p)
 
     CHECK_POINT();
     
-    for (i = 0; i < qxl->num_modes; i++) {
+    for (i = 0; i < qxl->num_modes; i++) 
+    {
 	struct qxl_mode *m = qxl->modes + i;
+
 	if (m->x_res == p->HDisplay &&
 	    m->y_res == p->VDisplay &&
 	    m->bits == pScrn->bitsPerPixel)
 	{
-	    return m;
+	    /* What QXL calls 16 bit is actually x1r5g5b515 */
+	    if (m->bits == 16) 
+	    {
+		if (pScrn->depth == 15)
+		    return i;
+	    }
+	    else if (pScrn->depth == m->bits)
+	    {
+		return i;
+	    }
 	}
     }
 
-    return NULL;	
+    return -1;
 }
 
 static ModeStatus
@@ -1182,6 +1215,7 @@ qxl_valid_mode(int scrn, DisplayModePtr p, Bool flag, int pass)
     ScrnInfoPtr pScrn = xf86Screens[scrn];
     qxl_screen_t *qxl = pScrn->driverPrivate;
     int bpp = pScrn->bitsPerPixel;
+    int mode_idx;
 
     /* FIXME: I don't think this is necessary now that we report the
      * correct amount of video ram?
@@ -1189,9 +1223,11 @@ qxl_valid_mode(int scrn, DisplayModePtr p, Bool flag, int pass)
     if (p->HDisplay * p->VDisplay * (bpp/8) > qxl->draw_area_size)
 	return MODE_MEM;
 
-    p->Private = (void *)qxl_find_native_mode(pScrn, p);
-    if (!p->Private)
-       return MODE_NOMODE;
+    mode_idx = qxl_find_native_mode (pScrn, p);
+    if (mode_idx == -1)
+	return MODE_NOMODE;
+
+    p->Private = (void *)(unsigned long)mode_idx;
 
     assert (((struct qxl_mode *)p->Private)->x_res == p->HDisplay);
     assert (((struct qxl_mode *)p->Private)->y_res == p->VDisplay);
@@ -1288,6 +1324,8 @@ qxl_pre_init(ScrnInfoPtr pScrn, int flags)
 	goto out;
     }
 
+    print_modes (qxl, scrnIndex);
+
     /* hate */
     qxl_unmap_memory(qxl, scrnIndex);
 
